<?php

/* This implementation will not support installation from remote source. Just
 * installation of already presented modules mainly to persistence preparation.
 */

class ModuleType
{
  const System = 0;
  const Normal = 1;
  const Application = 2;
}

class ModuleAction
{
  const Start = 0;
  const Stop = 1;
  const Install = 2;
  const Uninstall = 3;
  const Upgrade = 4;
  const Enable = 5;
  const Disable = 6;
}

class ModuleCondition
{
  const All = 0;
  const Enabled = 1;
  const NotEnabled = 2;
  const Installed = 3;
  const NotInstalled = 4;
  const Running = 5;
  const NotRunning = 6;
}

class AppModule
{
  public int $Id;
  public string $Name;
  public string $Title;
  public string $Version;
  public string $License;
  public string $Creator;
  public string $HomePage;
  public string $Description;
  public bool $Running;
  public bool $Enabled;
  public bool $Installed;
  public string $InstalledVersion;
  public int $Type;
  public array $Dependencies;
  public Database $Database;
  public Application $System;
  public AppModuleManager $Manager;
  public $OnChange;
  public array $Models;

  function __construct(Application $System)
  {
    $this->System = &$System;
    $this->Database = &$System->Database;
    $this->Installed = false;
    $this->Enabled = false;
    $this->Running = false;
    $this->HomePage = '';
    $this->License = '';
    $this->Version = '';
    $this->Creator = '';
    $this->Description = '';
    $this->Dependencies = array();
    $this->Type = ModuleType::Normal;
    $this->Models = array();
  }

  function Install(): void
  {
    if ($this->Installed) return;
    $List = array();
    $this->Manager->EnumDependenciesCascade($this, $List, array(ModuleCondition::NotInstalled));
    $this->Manager->Perform($List, array(ModuleAction::Install), array(ModuleCondition::NotInstalled));
    $this->DoInstall();
    $this->Installed = true;
    $this->InstalledVersion = $this->Version;
    $this->Manager->Modules[$this->Name] = $this;
  }

  function Uninstall(): void
  {
    if (!$this->Installed) return;
    $this->Stop();
    $this->Installed = false;
    $List = array();
    $this->Manager->EnumSuperiorDependenciesCascade($this, $List, array(ModuleCondition::Installed));
    $this->Manager->Perform($List, array(ModuleAction::Uninstall), array(ModuleCondition::Installed));
    $this->DoUninstall();
  }

  function Upgrade(): void
  {
    if (!$this->Installed) return;
    if ($this->InstalledVersion == $this->Version) return;
    $List = array();
    $this->Manager->EnumSuperiorDependenciesCascade($this, $List, array(ModuleCondition::Installed));
    $this->Manager->Perform($List, array(ModuleAction::Upgrade), array(ModuleCondition::Installed));
    $this->DoUpgrade();
  }

  function Reinstall(): void
  {
    $this->Uninstall();
    // TODO: Install also back dependecies
    $this->Install();
  }

  function Start(): void
  {
    if ($this->Running) return;
    if (!$this->Installed) return;
    $List = array();
    $this->Manager->EnumDependenciesCascade($this, $List, array(ModuleCondition::NotRunning));
    $this->Manager->Perform($List, array(ModuleAction::Start), array(ModuleCondition::NotRunning));
    $this->DoStart();
    $this->Running = true;
  }

  function Stop(): void
  {
    if (!$this->Running) return;
    $this->Running = false;
    $List = array();
    $this->Manager->EnumSuperiorDependenciesCascade($this, $List, array(ModuleCondition::Running));
    $this->Manager->Perform($List, array(ModuleAction::Stop), array(ModuleCondition::Running));
    $this->DoStop();
  }

  function Restart(): void
  {
    $this->Stop();
    $this->Start();
  }

  function Enable(): void
  {
    if ($this->Enabled) return;
    if (!$this->Installed) return;
    $List = array();
    $this->Manager->EnumDependenciesCascade($this, $List, array(ModuleCondition::NotEnabled));
    $this->Manager->Perform($List, array(ModuleAction::Enable), array(ModuleCondition::NotEnabled));
    $this->Enabled = true;
  }

  function Disable(): void
  {
    if (!$this->Enabled) return;
    $this->Stop();
    $this->Enabled = false;
    $List = array();
    $this->Manager->EnumSuperiorDependenciesCascade($this, $List, array(ModuleCondition::Enabled));
    $this->Manager->Perform($List, array(ModuleAction::Disable), array(ModuleCondition::Enabled));
  }

  protected function DoStart(): void
  {
  }

  protected function DoStop(): void
  {
  }

  protected function DoInstall(): void
  {
  }

  protected function DoUninstall(): void
  {
  }

  protected function DoUpgrade(): void
  {
  }

  function AddModel(Model $Model): void
  {
    $this->Models[get_class($Model)] = $Model;
  }
}

/* Manage installed modules */
class AppModuleManager
{
  public array $Modules;
  public System $System;
  public string $FileName;
  public string $ModulesDir;
  public $OnLoadModules;

  function __construct(System $System)
  {
    $this->Modules = array();
    $this->System = &$System;
    $this->FileName = dirname(__FILE__).'/../../Config/ModulesConfig.php';
    $this->ModulesDir = dirname(__FILE__).'/../../Modules';
  }

  function Perform(array $List, array $Actions, array $Conditions = array(ModuleCondition::All)): void
  {
    foreach ($List as $Index => $Module)
    {
      if (in_array(ModuleCondition::All, $Conditions) or
        ($Module->Running and in_array(ModuleCondition::Running, $Conditions)) or
        (!$Module->Running and in_array(ModuleCondition::NotRunning, $Conditions)) or
        ($Module->Installed and in_array(ModuleCondition::Installed, $Conditions)) or
        (!$Module->Installed and in_array(ModuleCondition::NotInstalled, $Conditions)) or
        ($Module->Enabled and in_array(ModuleCondition::Enabled, $Conditions)) or
        (!$Module->Enabled and in_array(ModuleCondition::NotEnabled, $Conditions)))
      {
        foreach ($Actions as $Action)
        {
          if ($Action == ModuleAction::Start) $Module->Start();
          if ($Action == ModuleAction::Stop) $Module->Stop();
          if ($Action == ModuleAction::Install) $Module->Install();
          if ($Action == ModuleAction::Uninstall) $Module->Uninstall();
          if ($Action == ModuleAction::Enable) $Module->Enable();
          if ($Action == ModuleAction::Disable) $Module->Disable();
          if ($Action == ModuleAction::Upgrade) $Module->Upgrade();
        }
      }
    }
  }

  function EnumDependenciesCascade(AppModule $Module, array &$List, array $Conditions = array(ModuleCondition::All))
  {
    foreach ($Module->Dependencies as $Dependency)
    {
      if (!array_key_exists($Dependency, $this->Modules))
        throw new Exception(sprintf(T('Module "%s" dependency "%s" not found'), $Module->Name, $Dependency));
      $DepModule = $this->Modules[$Dependency];
      if (in_array(ModuleCondition::All, $Conditions) or
        ($Module->Running and in_array(ModuleCondition::Running, $Conditions)) or
        (!$Module->Running and in_array(ModuleCondition::NotRunning, $Conditions)) or
        ($Module->Enabled and in_array(ModuleCondition::Enabled, $Conditions)) or
        (!$Module->Enabled and in_array(ModuleCondition::NotEnabled, $Conditions)) or
        ($Module->Installed and in_array(ModuleCondition::Installed, $Conditions)) or
        (!$Module->Installed and in_array(ModuleCondition::NotInstalled, $Conditions)))
      {
        array_push($List, $DepModule);
        $this->EnumDependenciesCascade($DepModule, $List, $Conditions);
      }
    }
  }

  function EnumSuperiorDependenciesCascade(AppModule $Module, array &$List, array $Conditions = array(ModuleCondition::All))
  {
    foreach ($this->Modules as $RefModule)
    {
      if (in_array($Module->Name, $RefModule->Dependencies) and
          (in_array(ModuleCondition::All, $Conditions) or
          ($Module->Running and in_array(ModuleCondition::Running, $Conditions)) or
          (!$Module->Running and in_array(ModuleCondition::NotRunning, $Conditions)) or
          ($Module->Enabled and in_array(ModuleCondition::Enabled, $Conditions)) or
          (!$Module->Enabled and in_array(ModuleCondition::NotEnabled, $Conditions)) or
          ($Module->Installed and in_array(ModuleCondition::Installed, $Conditions)) or
          (!$Module->Installed and in_array(ModuleCondition::NotInstalled, $Conditions))))
      {
        array_push($List, $RefModule);
        $this->EnumSuperiorDependenciesCascade($RefModule, $List, $Conditions);
      }
    }
  }

  function Start(): void
  {
    $this->LoadModules();
    if (file_exists($this->FileName)) $this->LoadState();
    $this->StartEnabled();
  }

  function StartAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Start));
  }

  function StartEnabled(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Start), array(ModuleCondition::Enabled));
  }

  function StopAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Stop));
  }

  function InstallAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Install));
    $this->SaveState();
  }

  function UninstallAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Uninstall));
    $this->SaveState();
  }

  function EnableAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Enable));
    $this->SaveState();
  }

  function DisableAll(): void
  {
    $this->Perform($this->Modules, array(ModuleAction::Disable));
    $this->SaveState();
  }

  function ModulePresent(string $Name): bool
  {
    return array_key_exists($Name, $this->Modules);
  }

  function ModuleEnabled(string $Name): bool
  {
    return array_key_exists($Name, $this->Modules) and $this->Modules[$Name]->Enabled;
  }

  function ModuleRunning(string $Name): bool
  {
    return array_key_exists($Name, $this->Modules) and $this->Modules[$Name]->Running;
  }

  function GetModule(string $Name): AppModule
  {
    return $this->Modules[$Name];
  }

  function SearchModuleById(int $Id): string
  {
    foreach ($this->Modules as $Module)
    {
      //DebugLog($Module->Name.' '.$Module->Id);
      if ($Module->Id == $Id) return $Module->Name;
    }
    return '';
  }

  function LoadState(): void
  {
    $ConfigModules = array();
    include($this->FileName);
    foreach ($ConfigModules as $Mod)
    {
      if (array_key_exists($Mod['Name'], $this->Modules))
      {
        $this->Modules[$Mod['Name']] = $this->Modules[$Mod['Name']];
        $this->Modules[$Mod['Name']]->Enabled = $Mod['Enabled'];
        $this->Modules[$Mod['Name']]->Installed = $Mod['Installed'];
        $this->Modules[$Mod['Name']]->InstalledVersion = $Mod['Version'];
      }
    }
  }

  function SaveState(): void
  {
    $Data = array();
    foreach ($this->Modules as $Module)
    {
      $Data[] = array('Name' => $Module->Name, 'Enabled' => $Module->Enabled,
        'Version' => $Module->Version, 'Installed' => $Module->Installed);
    }
    if (file_put_contents($this->FileName, "<?php \n\n\$ConfigModules = ".var_export($Data, true).";\n") === FALSE)
    {
      echo($this->FileName.' is not writeable.');
    }
  }

  function RegisterModule(AppModule $Module): void
  {
    $this->Modules[$Module->Name] = &$Module;
    $Module->Manager = &$this;
    $Module->OnChange = &$this->OnModuleChange;
  }

  function UnregisterModule(AppModule $Module): void
  {
    unset($this->Modules[array_search($Module, $this->Modules)]);
  }

  function LoadModulesFromDir(string $Directory): void
  {
    $List = scandir($Directory);
    foreach ($List as $Item)
    {
      if (is_dir($Directory.'/'.$Item) and ($Item != '.') and ($Item != '..') and ($Item != '.svn'))
      {
        include_once($Directory.'/'.$Item.'/'.$Item.'.php');
        $ModuleName = 'Module'.$Item;
        $this->RegisterModule(new $ModuleName($this->System));
      }
    }
  }

  function LoadModules(): void
  {
    if (is_array($this->OnLoadModules) and (count($this->OnLoadModules) == 2) and method_exists($this->OnLoadModules[0], $this->OnLoadModules[1]))
      call_user_func($this->OnLoadModules);
    else $this->LoadModulesFromDir($this->ModulesDir);
  }
}
